@@ -249,7 +249,7 @@ pub enum PollResult {
 /// ```
 ///
 /// This agent does not understand Candid, and only acts on byte buffers.
-#[derive(Clone, Debug)]
+#[derive(Clone)]
 pub struct Agent {
     nonce_factory: Arc<dyn NonceGenerator>,
     identity: Arc<dyn Identity>,
@@ -258,6 +258,14 @@ pub struct Agent {
     transport: Arc<dyn ReplicaV2Transport>,
 }
 
+impl fmt::Debug for Agent {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> Result<(), std::fmt::Error> {
+        f.debug_struct("Agent")
+            .field("ingress_expiry_duration", &self.ingress_expiry_duration)
+            .finish_non_exhaustive()
+    }
+}
+
 impl Agent {
     /// Create an instance of an [`AgentBuilder`] for building an [`Agent`]. This is simpler than
     /// using the [`AgentConfig`] and [`Agent::new()`].
@@ -516,10 +524,11 @@ impl Agent {
     pub async fn poll(
         &self,
         request_id: &RequestId,
-        effective_canister_id: &Principal,
+        effective_canister_id: Principal,
+        disable_range_check: bool,
     ) -> Result<PollResult, AgentError> {
         match self
-            .request_status_raw(request_id, *effective_canister_id)
+            .request_status_raw(request_id, effective_canister_id, disable_range_check)
             .await?
         {
             RequestStatusResponse::Unknown => Ok(PollResult::Submitted),
@@ -549,13 +558,17 @@ impl Agent {
     pub async fn wait<W: Waiter>(
         &self,
         request_id: RequestId,
-        effective_canister_id: &Principal,
+        effective_canister_id: Principal,
+        disable_range_check: bool,
         mut waiter: W,
     ) -> Result<Vec<u8>, AgentError> {
         waiter.start();
         let mut request_accepted = false;
         loop {
-            match self.poll(&request_id, effective_canister_id).await? {
+            match self
+                .poll(&request_id, effective_canister_id, disable_range_check)
+                .await?
+            {
                 PollResult::Submitted => {}
                 PollResult::Accepted => {
                     if !request_accepted {
@@ -587,6 +600,7 @@ impl Agent {
         &self,
         paths: Vec<Vec<Label>>,
         effective_canister_id: Principal,
+        disable_range_check: bool,
     ) -> Result<Certificate<'_>, AgentError> {
         let request = self.read_state_content(paths)?;
         let serialized_bytes = sign_request(&request, self.identity.clone())?;
@@ -594,10 +608,9 @@ impl Agent {
         let read_state_response: ReadStateResponse = self
             .read_state_endpoint(effective_canister_id, serialized_bytes)
             .await?;
-
         let cert: Certificate = serde_cbor::from_slice(&read_state_response.certificate)
             .map_err(AgentError::InvalidCborData)?;
-        self.verify(&cert)?;
+        self.verify(&cert, effective_canister_id, disable_range_check)?;
         Ok(cert)
     }
 
@@ -610,7 +623,13 @@ impl Agent {
     }
 
     /// Verify a certificate, checking delegation if present.
-    pub fn verify(&self, cert: &Certificate) -> Result<(), AgentError> {
+    /// Only passes if the certificate also has authority over the canister.
+    pub fn verify(
+        &self,
+        cert: &Certificate,
+        effective_canister_id: Principal,
+        disable_range_check: bool,
+    ) -> Result<(), AgentError> {
         let sig = &cert.signature;
 
         let root_hash = cert.tree.digest();
@@ -618,9 +637,9 @@ impl Agent {
         msg.extend_from_slice(IC_STATE_ROOT_DOMAIN_SEPARATOR);
         msg.extend_from_slice(&root_hash);
 
-        let der_key = self.check_delegation(&cert.delegation)?;
+        let der_key =
+            self.check_delegation(&cert.delegation, effective_canister_id, disable_range_check)?;
         let key = extract_der(der_key)?;
-
         let result = bls::core_verify(sig, &*msg, &*key);
         if result != bls::BLS_OK {
             Err(AgentError::CertificateVerificationFailed())
@@ -629,13 +648,33 @@ impl Agent {
         }
     }
 
-    fn check_delegation(&self, delegation: &Option<Delegation>) -> Result<Vec<u8>, AgentError> {
+    fn check_delegation(
+        &self,
+        delegation: &Option<Delegation>,
+        effective_canister_id: Principal,
+        disable_range_check: bool,
+    ) -> Result<Vec<u8>, AgentError> {
         match delegation {
             None => self.read_root_key(),
             Some(delegation) => {
                 let cert: Certificate = serde_cbor::from_slice(&delegation.certificate)
                     .map_err(AgentError::InvalidCborData)?;
-                self.verify(&cert)?;
+                self.verify(&cert, effective_canister_id, disable_range_check)?;
+                let canister_range_lookup = [
+                    "subnet".into(),
+                    delegation.subnet_id.clone().into(),
+                    "canister_ranges".into(),
+                ];
+                let canister_range = lookup_value(&cert, canister_range_lookup)?;
+                let ranges: Vec<(Principal, Principal)> =
+                    serde_cbor::from_slice(canister_range).map_err(AgentError::InvalidCborData)?;
+                if !disable_range_check
+                    && !principal_is_within_ranges(&effective_canister_id, &ranges[..])
+                {
+                    // the certificate is not authorized to answer calls for this canister
+                    return Err(AgentError::CertificateNotAuthorized());
+                }
+
                 let public_key_path = [
                     "subnet".into(),
                     delegation.subnet_id.clone().into(),
@@ -651,10 +690,13 @@ impl Agent {
         &self,
         canister_id: Principal,
         path: &str,
+        disable_range_check: bool,
     ) -> Result<Vec<u8>, AgentError> {
         let paths: Vec<Vec<Label>> = vec![vec!["canister".into(), canister_id.into(), path.into()]];
 
-        let cert = self.read_state_raw(paths, canister_id).await?;
+        let cert = self
+            .read_state_raw(paths, canister_id, disable_range_check)
+            .await?;
 
         lookup_canister_info(cert, canister_id, path)
     }
@@ -664,6 +706,7 @@ impl Agent {
         &self,
         canister_id: Principal,
         path: &str,
+        disable_range_check: bool,
     ) -> Result<Vec<u8>, AgentError> {
         let paths: Vec<Vec<Label>> = vec![vec![
             "canister".into(),
@@ -672,7 +715,9 @@ impl Agent {
             path.into(),
         ]];
 
-        let cert = self.read_state_raw(paths, canister_id).await?;
+        let cert = self
+            .read_state_raw(paths, canister_id, disable_range_check)
+            .await?;
 
         lookup_canister_metadata(cert, canister_id, path)
     }
@@ -682,11 +727,14 @@ impl Agent {
         &self,
         request_id: &RequestId,
         effective_canister_id: Principal,
+        disable_range_check: bool,
     ) -> Result<RequestStatusResponse, AgentError> {
         let paths: Vec<Vec<Label>> =
             vec![vec!["request_status".into(), request_id.to_vec().into()]];
 
-        let cert = self.read_state_raw(paths, effective_canister_id).await?;
+        let cert = self
+            .read_state_raw(paths, effective_canister_id, disable_range_check)
+            .await?;
 
         lookup_request_status(cert, request_id)
     }
@@ -699,6 +747,7 @@ impl Agent {
         request_id: &RequestId,
         effective_canister_id: Principal,
         signed_request_status: Vec<u8>,
+        disable_range_check: bool,
     ) -> Result<RequestStatusResponse, AgentError> {
         let _envelope: Envelope<ReadStateContent> =
             serde_cbor::from_slice(&signed_request_status).map_err(AgentError::InvalidCborData)?;
@@ -708,7 +757,7 @@ impl Agent {
 
         let cert: Certificate = serde_cbor::from_slice(&read_state_response.certificate)
             .map_err(AgentError::InvalidCborData)?;
-        self.verify(&cert)?;
+        self.verify(&cert, effective_canister_id, disable_range_check)?;
         lookup_request_status(cert, request_id)
     }
 
@@ -765,6 +814,14 @@ impl Agent {
     }
 }
 
+// Checks if a principal is contained within a list of principal ranges
+// A range is a tuple: (low: Principal, high: Principal), as described here: https://docs.dfinity.systems/spec/public/#state-tree-subnet
+fn principal_is_within_ranges(principal: &Principal, ranges: &[(Principal, Principal)]) -> bool {
+    ranges
+        .iter()
+        .any(|r| principal >= &r.0 && principal <= &r.1)
+}
+
 fn construct_message(request_id: &RequestId) -> Vec<u8> {
     let mut buf = vec![];
     buf.extend_from_slice(IC_REQUEST_DOMAIN_SEPARATOR);
@@ -1083,6 +1140,7 @@ pub struct UpdateCall<'agent> {
     agent: &'agent Agent,
     request_id: Pin<Box<dyn Future<Output = Result<RequestId, AgentError>> + Send + 'agent>>,
     effective_canister_id: Principal,
+    disable_range_check: bool,
 }
 
 impl fmt::Debug for UpdateCall<'_> {
@@ -1116,7 +1174,12 @@ impl UpdateCall<'_> {
             let request_id = _self.request_id.await?;
             _self
                 .agent
-                .wait(request_id, &_self.effective_canister_id, waiter)
+                .wait(
+                    request_id,
+                    _self.effective_canister_id,
+                    _self.disable_range_check,
+                    waiter,
+                )
                 .await
         }
         Box::pin(run(self, waiter))
@@ -1139,11 +1202,16 @@ pub struct UpdateBuilder<'agent> {
     pub arg: Vec<u8>,
     /// The Unix timestamp that the request will expire at.
     pub ingress_expiry_datetime: Option<u64>,
+    disable_range_check: bool,
 }
 
 impl<'agent> UpdateBuilder<'agent> {
     /// Creates a new query builder with an agent for a particular canister method.
     pub fn new(agent: &'agent Agent, canister_id: Principal, method_name: String) -> Self {
+        // When calling provisional_create_canister_with_cycles, every effective_canister_id is valid.
+        // Therefore we need to disable the check for valid canister_ranges in the certificate validation.
+        // More info: https://docs.dfinity.systems/spec/public/#http-effective-canister-id
+        let disable_range_check = method_name == "provisional_create_canister_with_cycles";
         Self {
             agent,
             effective_canister_id: canister_id,
@@ -1151,6 +1219,7 @@ impl<'agent> UpdateBuilder<'agent> {
             method_name,
             arg: vec![],
             ingress_expiry_datetime: None,
+            disable_range_check,
         }
     }
 
@@ -1218,6 +1287,7 @@ impl<'agent> UpdateBuilder<'agent> {
             agent: self.agent,
             request_id: Box::pin(request_id_future),
             effective_canister_id: self.effective_canister_id,
+            disable_range_check: self.disable_range_check,
         }
     }
 
